@@ -8,6 +8,7 @@ mod scroll_view;
 pub mod search;
 #[allow(dead_code)]
 mod select_view;
+mod status_bar;
 mod theme_view;
 pub mod toc;
 pub mod utils;
@@ -16,4 +17,5 @@ pub mod views {
     pub type ThemedView<T> = super::theme_view::ThemedView<T>;
     pub type RootLayout = super::root::RootLayout;
     pub type SelectView<T> = super::select_view::SelectView<T>;
+    pub type StatusBar = super::status_bar::StatusBar;
 }
