@@ -5,36 +5,35 @@ use select::document::Document;
 use serde::Deserialize;
 use serde_repr::Deserialize_repr;
 use std::{collections::HashMap, fmt::Display};
+use url::Url;
 
-use super::{parser::Parser, search::Namespace};
+use super::{language::Language, parser::Parser, search::Namespace};
 
-fn action_parse(params: Vec<(&str, String)>, url: String) -> Result<Response> {
-    Client::new()
-        .get(url)
-        .query(&[
-            ("action", "parse"),
-            ("format", "json"),
-            ("formatversion", "2"),
-        ])
-        .query(&params)
-        .send()
-        .context("failed sending the request")
-}
-
-#[derive(Debug, PartialEq, Eq, Clone, Copy)]
+#[derive(Debug, PartialEq, Eq, Clone)]
 pub enum ElementType {
     Text,
     Newline,
-    Link,
+    Link(Link),
     Header,
     Unsupported,
+
     ListMarker,
+    ListItemStart,
+    ListItemEnd,
+
+    DescriptionListTermStart,
+    DescriptionListTermEnd,
+    DescriptionListDescriptionStart,
+    DescriptionListDescriptionEnd,
+
+    DisambiguationStart,
+    DisambiguationEnd,
 }
 
 #[derive(Debug, Clone)]
 pub struct Element {
     id: usize,
-    kind: ElementType,
+    pub kind: ElementType,
     content: String,
     style: Style,
     width: usize,
@@ -64,10 +63,6 @@ impl Element {
         self.id
     }
 
-    pub fn kind(&self) -> ElementType {
-        self.kind
-    }
-
     pub fn content(&self) -> &str {
         &self.content
     }
@@ -88,14 +83,67 @@ impl Element {
     }
 }
 
-#[derive(Debug, Deserialize)]
+pub mod link_data {
+    use url::Url;
+
+    use crate::wiki::{language::Language, search::Namespace};
+
+    #[derive(Debug, Clone, PartialEq, Eq)]
+    pub struct InternalData {
+        pub namespace: Namespace,
+        pub page: String,
+        pub title: String,
+        pub endpoint: Url,
+        pub language: Language,
+        pub anchor: Option<AnchorData>,
+    }
+
+    #[derive(Debug, Clone, PartialEq, Eq)]
+    pub struct AnchorData {
+        pub anchor: String,
+        pub title: String,
+    }
+
+    #[derive(Debug, Clone, PartialEq, Eq)]
+    pub struct RedLinkData {
+        pub url: Url,
+        pub title: String,
+    }
+
+    #[derive(Debug, Clone, PartialEq, Eq)]
+    pub struct ExternalData {
+        pub url: Url,
+    }
+
+    #[derive(Debug, Clone, PartialEq, Eq)]
+    pub struct ExternalToInteralData {}
+}
+
+#[derive(Debug, Clone, PartialEq, Eq)]
+pub enum Link {
+    /// Interal link to another page in the same wiki
+    Internal(link_data::InternalData),
+    /// Anchor to a specific section in the current page
+    /// Note: this only corresponds to anchors on the current page. For anchors in another page on
+    /// the same wiki, `LinkType::Internal` is used
+    Anchor(link_data::AnchorData),
+    /// A special type of link that leads to an internal page that doesn't exist yet
+    RedLink(link_data::RedLinkData),
+    /// External link to a page at another website
+    External(link_data::ExternalData),
+    /// External link to an interal page in the same wiki
+    ExternalToInternal(link_data::ExternalToInteralData),
+}
+
+#[derive(Debug, Deserialize, Clone)]
 pub struct LanguageLink {
     #[serde(rename = "langname")]
-    name: String,
+    pub name: String,
     #[serde(rename = "lang")]
-    language: String,
-    autonym: String,
-    title: String,
+    pub language: Language,
+    pub autonym: String,
+    pub title: String,
+    pub url: Url,
 }
 
 #[derive(Debug, Deserialize)]
@@ -105,14 +153,6 @@ pub struct Category {
     hidden: Option<bool>,
 }
 
-#[derive(Debug, Deserialize)]
-pub struct Link {
-    #[serde(rename = "ns")]
-    namespace: Namespace,
-    title: String,
-    exists: bool,
-}
-
 #[derive(Debug, Deserialize)]
 pub struct Template {
     #[serde(rename = "ns")]
@@ -184,10 +224,10 @@ pub struct Article {
     title: String,
     pageid: usize,
     content: Option<Vec<Element>>,
-    language_links: Option<Vec<LanguageLink>>,
+    pub language: Language,
+    pub language_links: Option<Vec<LanguageLink>>,
     categories: Option<Vec<Category>>,
     categories_html: Option<String>,
-    links: Option<Vec<Link>>,
     templates: Option<Vec<Template>>,
     images: Option<Vec<String>>,
     external_links: Option<Vec<String>>,
@@ -208,7 +248,7 @@ pub struct Article {
 }
 
 impl Article {
-    pub fn builder() -> ArticleBuilder<NoPageID, NoPage, NoUrl> {
+    pub fn builder() -> ArticleBuilder<NoPageID, NoPage, NoEndpoint, NoLanguage> {
         ArticleBuilder::default()
     }
 
@@ -219,8 +259,20 @@ impl Article {
     pub fn sections(&self) -> Option<impl Iterator<Item = &Section>> {
         self.sections.as_ref().map(|x| x.iter())
     }
+
+    pub fn title(&self) -> &str {
+        &self.title
+    }
+
+    pub fn available_languages(&self) -> Option<usize> {
+        if let Some(ref links) = self.language_links {
+            return Some(links.len());
+        }
+        None
+    }
 }
 
+#[derive(Clone)]
 /// Which pieces of information to get about the article
 pub enum Property {
     /// Gives the parsed text of the wikitext
@@ -231,8 +283,6 @@ pub enum Property {
     Categories,
     /// Gives the HTML version of the categories
     CategoriesHTML,
-    /// Gives the interal links in the parsed wikitext
-    Links,
     /// Gives the templates in the parsed wikitext
     Templates,
     /// Gives the images in the parsed wikitext
@@ -276,7 +326,6 @@ impl Display for Property {
             Property::LangLinks => write!(f, "langlinks"),
             Property::Categories => write!(f, "categories"),
             Property::CategoriesHTML => write!(f, "categorieshtml"),
-            Property::Links => write!(f, "links"),
             Property::Templates => write!(f, "templates"),
             Property::Images => write!(f, "images"),
             Property::ExternalLinks => write!(f, "externallinks"),
@@ -306,52 +355,86 @@ pub struct Page(String);
 #[derive(Default)]
 pub struct NoPage;
 
-pub struct WithUrl(String);
+pub struct WithEndpoint(Url);
 #[derive(Default)]
-pub struct NoUrl;
+pub struct NoEndpoint;
+
+pub struct WithLanguage(Language);
+#[derive(Default)]
+pub struct NoLanguage;
 
 #[derive(Default)]
-pub struct ArticleBuilder<I, P, U> {
+pub struct ArticleBuilder<I, P, E, L> {
     pageid: I,
     page: P,
-    url: U,
+    endpoint: E,
+    language: L,
     revision: Option<usize>,
     redirects: Option<bool>,
     properties: Option<Vec<Property>>,
 }
 
-impl<U> ArticleBuilder<NoPageID, NoPage, U> {
+impl<E, L> ArticleBuilder<NoPageID, NoPage, E, L> {
     /// Parse content of this page
-    pub fn pageid(self, pageid: usize) -> ArticleBuilder<PageID, NoPage, U> {
+    pub fn pageid(self, pageid: usize) -> ArticleBuilder<PageID, NoPage, E, L> {
         ArticleBuilder {
             pageid: PageID(pageid),
             page: self.page,
-            url: self.url,
+            endpoint: self.endpoint,
             revision: self.revision,
             redirects: self.redirects,
             properties: self.properties,
+            language: self.language,
         }
     }
 
     /// Parse content of this page
-    pub fn page(self, page: impl Into<String>) -> ArticleBuilder<NoPageID, Page, U> {
+    pub fn page(self, page: impl Into<String>) -> ArticleBuilder<NoPageID, Page, E, L> {
         ArticleBuilder {
             pageid: self.pageid,
             page: Page(page.into()),
-            url: self.url,
+            endpoint: self.endpoint,
             revision: self.revision,
             redirects: self.redirects,
             properties: self.properties,
+            language: self.language,
         }
     }
 }
 
-impl<I, P> ArticleBuilder<I, P, NoUrl> {
-    pub fn url(self, url: impl Into<String>) -> ArticleBuilder<I, P, WithUrl> {
+impl<I, P, L> ArticleBuilder<I, P, NoEndpoint, L> {
+    pub fn url(self, url: impl Into<Url>) -> ArticleBuilder<I, P, WithEndpoint, L> {
         ArticleBuilder {
             pageid: self.pageid,
             page: self.page,
-            url: WithUrl(url.into()),
+            endpoint: WithEndpoint(url.into()),
+            revision: self.revision,
+            redirects: self.redirects,
+            properties: self.properties,
+            language: self.language,
+        }
+    }
+
+    pub fn endpoint(self, endpoint: Url) -> ArticleBuilder<I, P, WithEndpoint, L> {
+        ArticleBuilder {
+            pageid: self.pageid,
+            page: self.page,
+            endpoint: WithEndpoint(endpoint),
+            revision: self.revision,
+            redirects: self.redirects,
+            properties: self.properties,
+            language: self.language,
+        }
+    }
+}
+
+impl<I, P, E> ArticleBuilder<I, P, E, NoLanguage> {
+    pub fn language(self, language: Language) -> ArticleBuilder<I, P, E, WithLanguage> {
+        ArticleBuilder {
+            pageid: self.pageid,
+            page: self.page,
+            endpoint: self.endpoint,
+            language: WithLanguage(language),
             revision: self.revision,
             redirects: self.redirects,
             properties: self.properties,
@@ -359,7 +442,7 @@ impl<I, P> ArticleBuilder<I, P, NoUrl> {
     }
 }
 
-impl<I, P, U> ArticleBuilder<I, P, U> {
+impl<I, P, U, L> ArticleBuilder<I, P, U, L> {
     /// Revision ID, for `{{REVISIONID}}` and similar variables
     pub fn revision(mut self, revision: usize) -> Self {
         self.revision = Some(revision);
@@ -379,8 +462,25 @@ impl<I, P, U> ArticleBuilder<I, P, U> {
     }
 }
 
-impl<I, P> ArticleBuilder<I, P, WithUrl> {
+impl<I, P> ArticleBuilder<I, P, WithEndpoint, WithLanguage> {
     fn fetch_with_params(self, mut params: Vec<(&str, String)>) -> Result<Article> {
+        fn action_parse(params: Vec<(&str, String)>, endpoint: Url) -> Result<Response> {
+            Client::new()
+                .get(endpoint)
+                .query(&[
+                    ("action", "parse"),
+                    ("format", "json"),
+                    ("formatversion", "2"),
+                ])
+                .query(&params)
+                .send()
+                .map(|response| {
+                    debug!("response url: '{}'", response.url().as_str());
+                    response
+                })
+                .context("failed sending the request")
+        }
+
         if let Some(revision) = self.revision {
             params.push(("revid", revision.to_string()));
         }
@@ -389,7 +489,7 @@ impl<I, P> ArticleBuilder<I, P, WithUrl> {
             params.push(("redirects", redirects.to_string()));
         }
 
-        if let Some(prop) = self.properties {
+        if let Some(ref prop) = self.properties {
             let mut prop_str = String::new();
             for prop in prop {
                 prop_str.push('|');
@@ -398,14 +498,19 @@ impl<I, P> ArticleBuilder<I, P, WithUrl> {
             params.push(("prop", prop_str));
         }
 
-        let response = action_parse(params, self.url.0)?
+        let response = action_parse(params, self.endpoint.0.clone())?
             .error_for_status()
-            .context("recieved an error")?;
+            .context("the server returned an error")?;
 
         let res_json: serde_json::Value =
             serde_json::from_str(&response.text().context("failed reading the response")?)
-                .context("failed reading the response as json")?;
+                .context("failed interpreting the response as json")?;
 
+        self.serialize_result(res_json)
+            .context("failed serializing the returned response")
+    }
+
+    fn serialize_result(self, res_json: serde_json::Value) -> Result<Article> {
         let title = res_json
             .get("parse")
             .and_then(|x| x.get("title"))
@@ -427,8 +532,16 @@ impl<I, P> ArticleBuilder<I, P, WithUrl> {
             .map(|x| x.to_owned())
             .map(|x| {
                 x.into_iter()
-                    .filter_map(|x| serde_json::from_value(x).ok())
+                    .filter_map(|x| {
+                        serde_json::from_value(x)
+                            .map_err(|err| warn!("language_link parsing error: {:?}", err))
+                            .ok()
+                    })
                     .collect::<Vec<LanguageLink>>()
+            })
+            .map(|x| {
+                debug!("language_links: '{}'", x.len());
+                x
             });
 
         let categories = res_json
@@ -448,17 +561,6 @@ impl<I, P> ArticleBuilder<I, P, WithUrl> {
             .and_then(|x| x.as_str())
             .map(|x| x.to_owned());
 
-        let links = res_json
-            .get("parse")
-            .and_then(|x| x.get("links"))
-            .and_then(|x| x.as_array())
-            .map(|x| x.to_owned())
-            .map(|x| {
-                x.into_iter()
-                    .filter_map(|x| serde_json::from_value(x).ok())
-                    .collect::<Vec<Link>>()
-            });
-
         let templates = res_json
             .get("parse")
             .and_then(|x| x.get("templates"))
@@ -531,7 +633,16 @@ impl<I, P> ArticleBuilder<I, P, WithUrl> {
             .get("parse")
             .and_then(|x| x.get("text"))
             .and_then(|x| x.as_str())
-            .and_then(|x| Parser::parse_document(x, &title, sections.as_ref()).ok());
+            .and_then(|x| {
+                Parser::parse_document(
+                    x,
+                    &title,
+                    sections.as_ref(),
+                    self.endpoint.0.clone(),
+                    self.language.0.clone(),
+                )
+                .ok()
+            });
 
         let revision_id = res_json
             .get("parse")
@@ -578,10 +689,10 @@ impl<I, P> ArticleBuilder<I, P, WithUrl> {
             title,
             pageid,
             content,
+            language: self.language.0,
             language_links,
             categories,
             categories_html,
-            links,
             templates,
             images,
             external_links,
@@ -603,14 +714,14 @@ impl<I, P> ArticleBuilder<I, P, WithUrl> {
     }
 }
 
-impl ArticleBuilder<PageID, NoPage, WithUrl> {
+impl ArticleBuilder<PageID, NoPage, WithEndpoint, WithLanguage> {
     pub fn fetch(self) -> Result<Article> {
         let param = vec![("pageid", self.pageid.0.to_string())];
         self.fetch_with_params(param)
     }
 }
 
-impl ArticleBuilder<NoPageID, Page, WithUrl> {
+impl ArticleBuilder<NoPageID, Page, WithEndpoint, WithLanguage> {
     pub fn fetch(self) -> Result<Article> {
         let param = vec![("page", self.page.0.to_string())];
         self.fetch_with_params(param)
